#include "distributions.h"

#include <iostream>
#include <fstream>
#include <cmath>
#include <boost/math/distributions/normal.hpp>
#include <cstdlib>
#include <ctime>

#define CRITVALSAMPLESIZE 200

extern long double chisqcval;


int compareLongDouble(const void *p1,const void *p2);
using namespace std;

extern boost::math::normal stdnorm;


inline long double pnorm(long double x){
  return erfc((-x)/sqrt(2))/2;
}

inline long double qnorm(long double p){
  return quantile(stdnorm,p);
}

inline long double rnorm(){
  return quantile(stdnorm,((long double)rand())/RAND_MAX);
}

void getpos(unsigned int n,const long double *list,unsigned int numps,const unsigned int *pos,long double *to);

const char *const normalvalidparnames[]={"mu","sigma","VaR@"};

class NormalDistribution: public Distribution{
protected:
  static const char* distnm;
  long double mu;
  long double sigma;
  char fixedmu;
  char fixedsig;
  unsigned int mainpar;
  long double VaR_at;
public:
  NormalDistribution(const NormalDistribution &orig):mu(orig.mu),sigma(orig.sigma),fixedmu(orig.fixedmu),fixedsig(orig.fixedsig),mainpar(orig.mainpar),VaR_at(orig.VaR_at){};
  NormalDistribution(long double m,long double s):mu(m),sigma(s),fixedmu(0),fixedsig(0),mainpar(0),VaR_at(0){};
  NormalDistribution():mu(0),sigma(1),fixedmu(0),fixedsig(0),mainpar(0),VaR_at(0){};
  ~NormalDistribution(){};
  unsigned int numpars() const{return 2;};
  unsigned int numvalidpars() const{return 3;};
  const char *const*validparnames() const{return normalvalidparnames;};
  long double dist(long double x) const;
  long double survive(long double x) const;
  long double quant(long double p) const;
  long double random() const;  
  long double parval() const;
  long double parval(unsigned int i,long double at) const;
  void setfixed(unsigned int i,long double val);
  void unfix(unsigned int i);
  void setmain(unsigned int i,long double at);
  long double transform_data(long double data) const{return data;};
  long double untransform_data(long double data) const{return data;};
  long double get_critval(long double size,AdequacyStat stat)  const;
  long double adjustADstat(unsigned int samplesize) const;
  unsigned int numestpars() const;
  Distribution* copy() const;
  void estimate_data(unsigned int n,const long double *data);
  void estimate_data(unsigned int n,const long double *data,const unsigned int *smp);
  void estimate_data(unsigned int c,const unsigned int *counts,const long double *cutoffs);//for grouped data
  const char *distname() const{return distnm;};
  void print() const;
};

unsigned int NormalDistribution::numestpars() const{
  unsigned int ans=2;
  if(fixedmu){
    ans--;
  };
  if(fixedsig){
    ans--;
  };
  return ans;
};


class LogNormalDistribution: public NormalDistribution{
  static const char* distnm;
public:
  LogNormalDistribution(const LogNormalDistribution &orig):NormalDistribution(orig){};
  LogNormalDistribution(long double m,long double s):NormalDistribution(m,s){};
  LogNormalDistribution():NormalDistribution(){};
  long double transform_data(long double data) const;
  long double untransform_data(long double data) const;
  void print() const;
  Distribution* copy() const;
  const char *distname() const{return distnm;};
  long double parval() const;  //Other functions are the same as for normal distribution.
};


const char *LogNormalDistribution::distnm="log-normal";
const char *NormalDistribution::distnm="normal";

Distribution* LogNormalDistribution::copy() const{
  return new LogNormalDistribution(*this);
};


long double CalculateADstat(unsigned int n,const long double *data, const Distribution &dist);

long double  NormalDistribution::get_critval(long double size,AdequacyStat stat) const{
  //perform 10000 simulations with CRITVALSAMPLESIZE data points each to get
  //distribution of critical value.
  //This function isn't very good.
  //I should improve it later.
  long double adval[10000];
  long double data[CRITVALSAMPLESIZE];
  for(unsigned int i=0;i<10000;i++){
    for(unsigned int j=0;j<CRITVALSAMPLESIZE;j++){
      data[j]=this->random();
    };    
    qsort(data,CRITVALSAMPLESIZE,sizeof(long double),compareLongDouble);
    adval[i]=(*stat)(CRITVALSAMPLESIZE,data,*this);
//    if(i>9790){
//      cout<<adval[i]<<"  ";
//    };
//    if(i>9790 && !(i%10)){
//      cout<<"\n";
//    };
  };
  unsigned int pos=(long double)10000*(1-size);
  if(pos>9999){
    pos=9999;
  };
  long double ans=0;
  getpos(10000,adval,1,&pos,&ans);
  ans*=this->adjustADstat(CRITVALSAMPLESIZE);
  //  cout<<pos<<"\t"<<ans<<"\n";
  return(ans);
};



long double NormalDistribution::adjustADstat(unsigned int samplesize) const{
  if(!fixedmu&&!fixedsig){
    return 1+0.75/samplesize+2.25/(samplesize*samplesize);
  }else{
    return 1;
  };
};

Distribution* NormalDistribution::copy() const{
  return new NormalDistribution(*this);
};

long double NormalDistribution::dist(long double x) const{
  //  cout<<x<<"  "<<mu<<"  "<<sigma<<"\n";
  return erfc((mu-x)/sqrt(2)/sigma)/2;  
};

long double NormalDistribution::survive(long double x) const{
  //  cout<<x<<"  "<<mu<<"  "<<sigma<<"\n";
  return erfc((x-mu)/sqrt(2)/sigma)/2;  
};

long double NormalDistribution::quant(long double p) const{
  return quantile(stdnorm,p)*sigma+mu;
};

long double NormalDistribution::random() const{
  return quantile(stdnorm,((long double)rand())/RAND_MAX)*sigma+mu;
};  




long double NormalDistribution::parval() const{
  switch(mainpar){
  case 0:
    return mu;
  case 1:
    return sigma;
  default:
    return this->quant(VaR_at);
  };
};

void NormalDistribution::setfixed(unsigned int i,long double val){
  switch(i){
  case 0:
    mu=val;
    fixedmu=1;
    break;
  case 1:
    sigma=val;
    fixedsig=1;
    break;
  default:
    cout<<"Fixing VaR not currently possible for normal distribution.\n";
    exit(0);
  };
};

void NormalDistribution::unfix(unsigned int i){
  switch(i){
  case 0:
    fixedmu=0;
    break;
  case 1:
    fixedsig=0;
    break;
  default:
    break;
    //VaR can't be fixed anyway.
  };
};


long double NormalDistribution::parval(unsigned int i,long double at) const{
  switch(i){
  case 0:
    return mu;
  case 1:
    return sigma;
  default:
    return this->quant(at);
  };
};

void NormalDistribution::setmain(unsigned int i,long double at){
  mainpar=i;
  if(i==2){
    VaR_at=at;
  };
};



void NormalDistribution::estimate_data(unsigned int n,const long double *data){
  long double m=0;
  long double ss=0;
  for(unsigned int i=0;i<n;i++){
    m+=*(data+i);
    ss+=*(data+i)*(*(data+i));
  };
  m/=n;
  if(!fixedmu){
    mu=m;
  };
  ss/=n;//Use the MLE, not the unbiassed estimator.
  if(!fixedsig){
    sigma=sqrt(ss-m*m+(mu-m)*(mu-m));
  };
};

void NormalDistribution::estimate_data(unsigned int n,const long double *data,const unsigned int *smp){
  long double m=0;
  long double ss=0;
  unsigned int tot=0;
  for(unsigned int i=0;i<n;i++){
    tot+=*(smp+i);
    m+=*(data+i)*(*(smp+i));
    ss+=*(data+i)*(*(data+i))*(*(smp+i));
  };
  m/=tot;
  if(!fixedmu){
    mu=m;
  };
  ss/=tot;//Use the MLE, not the unbiassed estimator.
  if(!fixedsig){
    sigma=sqrt(ss-m*m+(mu-m)*(mu-m));
  };
};

void NormalDistribution::estimate_data(unsigned int c,const unsigned int *counts,const long double *cutoffs){
  //Unfinished
  cout<<"Normal not yet available for grouped data.\n";
  exit(0);
};

long double LogNormalDistribution::transform_data(long double data) const{
  return(log(data));
};

long double LogNormalDistribution::untransform_data(long double data) const{
  return(exp(data));
};


void NormalDistribution::print() const{
  cout<<"normal distribution.\n\n";
  cout<<"mu="<<mu;
  if(fixedmu){
    cout<<"\t fixed";
  };
  cout<<"\nsigma="<<sigma;
  if(fixedsig){
    cout<<"\t fixed";
  };
  cout<<"\nParameter of interest: ";
  switch(mainpar){
  case 0:
    cout<<"mu\n";
    return;
  case 1:
    cout<<"sigma\n";
    return;
  case 2:
    cout<<"VaR at "<<VaR_at<<" level.\n";
    return;
  default:
    cout<<"No main parameter selected. This should not happen.\n";
  };
};


void LogNormalDistribution::print() const{
  cout<<"log-normal distribution.\n\n";
  cout<<"mu="<<mu;
  if(fixedmu){
    cout<<"\t fixed";
  };
  cout<<"\nsigma="<<sigma;
  if(fixedsig){
    cout<<"\t fixed";
  };
  cout<<"\nParameter of interest: ";
  switch(mainpar){
  case 0:
    cout<<"mu\n";
    return;
  case 1:
    cout<<"sigma\n";
    return;
  case 2:
    cout<<"VaR at "<<VaR_at<<" level.\n";
    return;
  default:
    cout<<"No main parameter selected. This should not happen.\n";
  };
};

long double LogNormalDistribution::parval() const{
  switch(mainpar){
  case 0:
    return mu;
  case 1:
    return sigma;
  default:
    return exp(this->quant(VaR_at));
  };
};


const char* const paretovalidparnames[]={"alpha","min","VaR@"};

class ParetoType1Distribution: public Distribution{
protected:
  static const char* distnm;
  long double alpha;
  long double min;
  char fixedalpha;
  char fixedmin;
  unsigned int mainpar;
  long double VaR_at;
public:
  ParetoType1Distribution(const ParetoType1Distribution &orig):alpha(orig.alpha),min(orig.min),fixedalpha(orig.fixedalpha),fixedmin(orig.fixedmin),mainpar(orig.mainpar),VaR_at(orig.VaR_at){};
  ParetoType1Distribution(long double al,long double mn):alpha(al),min(mn),fixedalpha(0),fixedmin(0),mainpar(0),VaR_at(0){};
  ParetoType1Distribution():alpha(2),min(1),fixedalpha(0),fixedmin(0),mainpar(0),VaR_at(0){};
  ~ParetoType1Distribution(){};
unsigned int numpars() const{return 2;};
  unsigned int numvalidpars() const{return 3;};
  const char *const*validparnames() const{return paretovalidparnames;};
  long double dist(long double x) const;
  long double survive(long double x) const;
  long double quant(long double p) const;
  long double random() const;  
  long double parval() const;
  long double parval(unsigned int i,long double at) const;
  void setfixed(unsigned int i,long double val);
  void unfix(unsigned int i);
  void setmain(unsigned int i,long double at);
  long double transform_data(long double data) const{return data;};
  long double untransform_data(long double data) const{return data;};
  long double get_critval(long double size,AdequacyStat stat)  const;
  long double adjustADstat(unsigned int samplesize) const;
  unsigned int numestpars() const;
  Distribution* copy() const;
  void estimate_data(unsigned int n,const long double *data);
  void estimate_data(unsigned int n,const long double *data,const unsigned int *smp);
  void estimate_data(unsigned int c,const unsigned int *counts,const long double *cutoffs);//for grouped data
  void print() const;
  const char *distname() const{return distnm;};
};

unsigned int ParetoType1Distribution::numestpars() const{
  unsigned int ans=2;
  if(fixedmin){
    ans--;
  };
  if(fixedalpha){
    ans--;
  };
  return ans;
};

const char *ParetoType1Distribution::distnm="normal";


long double ParetoType1Distribution::dist(long double x) const{
  return 1-pow(min/x,alpha);
};

long double ParetoType1Distribution::survive(long double x) const{
  return pow(min/x,alpha);
};

long double ParetoType1Distribution::quant(long double p) const{
  return min/pow(1-p,alpha);
};

long double ParetoType1Distribution::random() const{
  long double unif=0;
  while(unif==0){
    unif=((long double)rand())/RAND_MAX;
  };
  return min/pow(unif,alpha);
};  

long double ParetoType1Distribution::parval() const{
  switch(mainpar){
  case 0:
    return alpha;
  case 1:
    return min;
  default:
    return this->quant(VaR_at);
  };
};

long double ParetoType1Distribution::parval(unsigned int i,long double at) const{
  switch(i){
  case 0:
    return alpha;
  case 1:
    return min;
  default:
    return this->quant(at);
  };
};

void ParetoType1Distribution::setfixed(unsigned int i,long double val){
  switch(i){
  case 0:
    alpha=val;
    fixedalpha=1;
    break;
  case 1:
    if(val>0){
      min=val;
      fixedmin=1;
    }else{
      cout<<"Minimum for Pareto distribution must be positive. Ignoring fixed value.\n";
    };
    break;
  default:
    cout<<"Fixing VaR not currently possible for Pareto distribution.\n";
    exit(0);
  };
};

void ParetoType1Distribution::unfix(unsigned int i){
  switch(i){
  case 0:
    fixedalpha=0;
    break;
  case 1:
    fixedmin=0;
    break;
  default:
    break;
    //VaR can't be fixed anyway.
  };
};

void ParetoType1Distribution::setmain(unsigned int i,long double at){
  mainpar=i;
  if(i==2){
    VaR_at=at;
  };
};

void ParetoType1Distribution::estimate_data(unsigned int n,const long double *data){
  //density is alpha x^{-(alpha+1)} log-likelihood is log(alpha)-(alpha+1)log(x)
  //derivative wrt alpha is therefore 1/alpha-log(x)
  //estimate is alpha is the inverse of the average of log(x).
  if(fixedmin){
    for(unsigned int i=0;i<n;i++){
      if(*(data+i)<min){
	cerr<<"Data point below minimum specified value for Type I Pareto distribution.\n";
	cerr<<"Rerun with new minimum specified\n";
	exit(0);
      };
    };
  }else{
    if(*data<=0){
      cout<<"Only positive values are possible for the Pareto distribution.\n";
      cout<<"Exiting.\n";
      exit(0);
    };
    min=*(data);
    for(unsigned int i=1;i<n;i++){
      if(*(data+i)<min){
	if(*(data+i)<=0){
	  cout<<"Only positive values are possible for the Pareto distribution.\n";
	  cout<<"Exiting.\n";
	  exit(0);
	};
	min=*(data+i);
      };
    };    
  };
  if(!fixedalpha){
    long double slx=0;
    unsigned int x=0;
    for(unsigned int i=0;i<n;i++){
      if(*(data+i)>min){//truncate at min.
	slx+=log(*(data+i));
	x++;
      };
    };
    slx-=x*log(min);
    alpha=x/slx;
  };
};

void ParetoType1Distribution::estimate_data(unsigned int n,const long double *data,const unsigned int *smp){
  //density is alpha x^{-(alpha+1)} log-likelihood is log(alpha)-(alpha+1)log(x)
  //derivative wrt alpha is therefore 1/alpha-log(x)
  //estimate is alpha is the inverse of the average of log(x).
  if(fixedmin){
    for(unsigned int i=0;i<n;i++){
      if(*(data+i)<min&&*(smp+i)){
	cerr<<"Data point below minimum specified value for Type I Pareto distribution.\n";
	cerr<<"Rerun with new minimum specified\n";
	exit(0);
      };
    };
  }else{
    if(*data<=0&&*smp){
      cout<<"Only positive values are possible for the Pareto distribution.\n";
      cout<<"Exiting.\n";
      exit(0);
    };
    min=*(data);
    for(unsigned int i=1;i<n;i++){
      if(*(data+i)<min&&*(smp+i)){
	if(*(data+i)<=0){
	  cout<<"Only positive values are possible for the Pareto distribution.\n";
	  cout<<"Exiting.\n";
	  exit(0);
	};
	min=*(data+i);
      };
    };    
  };
  if(!fixedalpha){
    unsigned int x=0;
    long double slx=0;
    for(unsigned int i=0;i<n;i++){
      if(*(data+i)>=min){
	slx+=log(*(data+i))*(*(smp+i));
	x+=(*(smp+i));
      };
    };
    slx-=x*log(min);
    alpha=x/slx;
  };
};

void ParetoType1Distribution::estimate_data(unsigned int c,const unsigned int *counts,const long double *cutoffs){
  unsigned int tot=0;
  for(unsigned int i=0;i<c;i++){
    tot+=*(counts+i);
  };
  if(fixedalpha){
    min=pow(pow(*cutoffs,alpha)*(1-(long double )(*counts))/tot,1/alpha);
  }else if(fixedmin){
    alpha=2;
    long double dla=10;
    while(abs(dla)>1e-6){
      long double difz=(pow(*cutoffs,alpha)-pow(min,alpha));
      long double lr=log(min/(*cutoffs));
      dla=(tot-*counts)*log(min)-(*counts)*lr*pow(min,alpha)/difz-*(counts+c-1)*log(*(cutoffs+c-2));
      long double d2la=-((*counts)*lr*lr*(pow(*cutoffs*min,alpha))/(difz*difz));




      for(unsigned int i=1;i<c-1;i++){
	long double diff=pow(*(cutoffs+i),alpha)-pow(*(cutoffs+i-1),alpha);
	lr=log(*(cutoffs+i)/(*(cutoffs+i-1)));
	dla+=(*(counts+i))*((pow(*(cutoffs+i-1),alpha)*log(*(cutoffs+i))-pow(*(cutoffs+i),alpha)*log(*(cutoffs+i-1))))/diff;
	d2la-=*(counts+i)*(pow(*(cutoffs+i)*(*(cutoffs+i-1)),alpha)*lr*lr)/(diff*diff);
      };
      //            cout<<dla<<" "<<d2la<<"\n";
      long double move=dla/d2la;
      if(move<alpha){
	alpha-=move;
      }else{
	alpha/=3;
      };
	//	cout<<"alpha="<<alpha<<"\n";
    };    
  }else{
    alpha=2;
    min=*cutoffs;
    if(min<=0){
      min=0.5;
    };
    long double dla=10;
    while(abs(dla)>1e-6){
      long double lr=log(min/(*cutoffs));
      dla=(tot-*counts)*log(min)-*(counts+c-1)*log((*(cutoffs+c-2)));
      long double d2la=0;


      for(unsigned int i=1;i<c-1;i++){
	long double diff=pow(*(cutoffs+i),alpha)-pow(*(cutoffs+i-1),alpha);
	lr=log(*(cutoffs+i)/(*(cutoffs+i-1)));
	dla+=(*(counts+i))*((pow(*(cutoffs+i-1),alpha)*log(*(cutoffs+i))-pow(*(cutoffs+i),alpha)*log(*(cutoffs+i-1))))/diff;
	d2la-=*(counts+i)*(pow(*(cutoffs+i)*(*(cutoffs+i-1)),alpha)*lr*lr)/(diff*diff);
      };
      //                  cout<<dla<<" "<<d2la<<"\n";
      alpha-=dla/d2la;
      if(alpha<0){
	long double alphaold=alpha+dla/d2la;
	alpha=alphaold/2;
      };
      //      		cout<<"alpha="<<alpha<<"\n";
    };
    min=pow(pow(*cutoffs,alpha)*(1-((long double )(*counts))/tot),1/alpha);
  };
};//for grouped data

long double ParetoType1Distribution::get_critval(long double size,AdequacyStat stat)  const{
  long double adval[10000];
  long double data[CRITVALSAMPLESIZE];
  for(unsigned int i=0;i<10000;i++){
    for(unsigned int j=0;j<CRITVALSAMPLESIZE;j++){
      data[j]=this->random();
    };    
    qsort(data,CRITVALSAMPLESIZE,sizeof(long double),compareLongDouble);
    adval[i]=(*stat)(CRITVALSAMPLESIZE,data,*this);
  };
  unsigned int pos=(long double)10000*(1-size);
  if(pos>9999){
    pos=9999;
  };
  long double ans=0;
  getpos(10000,adval,1,&pos,&ans);
  ans*=this->adjustADstat(CRITVALSAMPLESIZE);
  //  cout<<pos<<"\t"<<ans<<"\n";
  return(ans);
};

long double ParetoType1Distribution::adjustADstat(unsigned int samplesize) const{
  return 1;
  //No adjustment. Perhaps should check this.
};

Distribution* ParetoType1Distribution::copy() const{
  return new ParetoType1Distribution(*this);
};

void ParetoType1Distribution::print() const{
  cout<<"Pareto distribution.\n\n";
  cout<<"alpha="<<alpha;
  if(fixedalpha){
    cout<<"\t fixed";
  };
  cout<<"\nmin="<<min;
  if(fixedmin){
    cout<<"\t fixed";
  };
  cout<<"\nParameter of interest: ";
  switch(mainpar){
  case 0:
    cout<<"alpha\n";
    return;
  case 1:
    cout<<"min\n";
    return;
  case 2:
    cout<<"VaR at "<<VaR_at<<" level.\n";
    return;
  default:
    cout<<"No main parameter selected. This should not happen.\n";
  };
};




class DistributionlistElement{
  const char* name;
  Distribution &dist;
  DistributionlistElement *next;
  DistributionlistElement();
public:
  DistributionlistElement(const char *nm,Distribution &dst);
  void insert(const char *nm,Distribution &dst);
  Distribution* match(const char *nm);
};

Distribution* DistributionlistElement::match(const char *nm){
  if(!strcmp(name,nm)){
    return &dist;
  }else if(next==NULL){
    return NULL;
  }else{
    return next->match(nm);
  };
};


DistributionlistElement::DistributionlistElement(const char *nm,Distribution &dst):name(nm),dist(dst),next(NULL){
};

void DistributionlistElement::insert(const char *nm,Distribution &dst){
  DistributionlistElement*nx=next;
  next=new DistributionlistElement(nm,dst);
  next->next=nx;
};

DistributionlistElement *distributions=NULL;

void set_up_distributions(){
  distributions=new DistributionlistElement("normal",*(new NormalDistribution()));
  distributions->insert("log_normal",*(new LogNormalDistribution()));
  distributions->insert("pareto",*(new ParetoType1Distribution()));
};

Distribution& read_distribution(int argc,char *argv[],unsigned int &start){
  Distribution* orig=(distributions->match(argv[start]));
  if(orig==NULL){
    //could not find matching distibution
    cout<<"Distribution \""<<argv[start]<<"\" not currently supported.\n";
    exit(0);
  };
  Distribution *ans=orig->copy();//Not freed at exit
  for(start++;start<(unsigned)argc;start++){
    char *at=NULL;
    char *eq=NULL;
    unsigned int percent=0;
    char* name=new char[strlen(argv[start])+1];
    strcpy(name,argv[start]);
    unsigned int len=strlen(name);
    for(unsigned int i=0;i<len;i++){
      if(*(name+i)=='@'){
	*(name+i)='\0';
	at=(name+i+1);
      }else if(*(name+i)=='%'){
	*(name+i)='\0';
	percent=1;
      }else if(*(name+i)=='='){
	*(name+i)='\0';
	eq=(name+i+1);	
      };
    };
    long double val=0;
    if(eq!=NULL){
      val=atof(eq);
    };
    long double atval=0;
    if(at!=NULL){
      atval=atof(at);
      if(eq!=NULL&&eq<at){
	*eq='\0';
	*(eq-1)='@';
      }else{
	*at='\0';
	*(at-1)='@';
      };
    };
    unsigned int i=0;
    for(;i<ans->numvalidpars()&&strcmp(name,*(ans->validparnames()+i));i++);
    if(i==ans->numvalidpars()){
      delete[] name;
      return *ans;
    };
    if(eq!=NULL){
      ans->setfixed(i,val);
      //      cout<<"set parameter "<<i<<" to value "<<val<<".\n";
      //      ans->print();
      if(percent){
	ans->unfix(i);
      };
    }else{
      ans->setmain(i,atval);
    };   
    delete[] name;
  };
  return *ans;
};

